 runinference.accurate<-function(M, maxn=1, 
                                 samples=1e15, #this is needed for the tests
                                 rules=c(1,2,3),verbose=0) {
  #Function the runs the learning inference on a model using accurate covariance data.
  #
  #INPUT:
  # M           - Model from randomModel.
  # maxn        - The maximum size of intervention set.
  # rules       - Which of the rules 1,2,3 should be used.
  # verbose     - Amount of printing.

  #OUTPUT:
  # list with 
  # $M                - Model inputted, the correct model.
  # $pc_significance  - What significance level was used for PC independence tests.
  # $pc_maxsetsize    - Maximum cond. set size for pc-skeleton.
  # $pind             - Independence p-level thresold.
  # $pdep             - Dependence thresold used.
  # $samples          - The number of samples used. For tests.
  # $sampling         - Whether data was sampled or not. Here FALSE.
  # $rules            - Vector of indexes of which faithfulness rules used. 
  # $Ei               - Matrix of experiments. Experiments are columns.
  # $extra_zero       - What is the largest element marked as zero by Faithfullness rules.
  # $extra_determined - What is the largest error if element is marked identified?
  #                     Only considers elements put determined when PC is not satisf. for the pair.
  # $PCs              - Vector of PCs satisfied after each experiment.
  # $Rs               - Vector of QR-ranks after each experiment.
  # $B                - The learned matrix B. Should match canonical($M)$B.
  # $nexp             - The number of experiments conducted.


  n<-sum(!M$hidden)

  #parameters for pc and ind and dep thresolds
  R<-list() #result list
  #parameters for pc and ind and dep thresolds
  R$pc_significance <- 0.05
  R$pc_maxsetsize <- NA
  R$pind <- 0.05
  R$pdep <- 0.05

  #finding the canonical model for checks
  C<-canonical(M)

  #these are the true zeros
  truezeros <- abs( C$B ) < 0.01

  #these keeps track of the zeros, to see which new ones are found
  zero <- array(FALSE,dim(C$B))

  #index for the experiments
  ei <- 0

  #was inference correct
  inferred <- FALSE

  #structure for the results
  R$M<-M#the original


  R$samples<-samples
  R$sampling<-FALSE #no sampling
  R$rules<-rules


  R$PCs<-c() #pair conditions satisfied
  R$Rs<-c() #ranks
  R$Ei <- array( 0, c(n,0) )

  R$extra_zero<-(-Inf) #marks the maximum element put to zero
  R$extra_determined<-(-Inf) #marks the biggest difference when determining zeros

  #this mask is used the ignore the diagonal in various places
  mask<-array(TRUE,c(n,n));
  diag(mask)<-FALSE

  eqs<-NULL #structure containing the equations

  #designing here all possible experiments
  experiments<-designExperiments(n,maxn)


  while ( !inferred ) {
    ei <- ei + 1

    if ( verbose > 0 ) cat('Choosing the next experiment...\n')

    ex_index <- chooseExperiment( eqs, experiments ) 
    e<-experiments[,ex_index]
    experiments[,ex_index]<-NA #not doing that experiment again

    if ( verbose > 0 ) cat('Next experiment: J=(', which(e>0),')\n')

    e <- experiment( M, e )
    attr(e$Cx,'N')<-samples #putting here in the samples

    R$Ei<-cbind(R$Ei,e$e)

    #add the experimental things to the list of experimental data
    if ( verbose > 0 ) cat('Adding equations...\n')
    eqs<-addEquations( e$e, e$Cx, eqs )


    if ( length(rules) > 0 ) {
      #keeping book how many times each element is put to zero
      Z<-array(0,c(n,n))

      if ( any(rules == 1 ) ) {
        pc_result <- pc_skeleton( e$Cx, samples,
                                  significance=R$pc_significance,
                                  maxsetsize=R$pc_maxsetsize)

        if ( verbose > 0 ) { cat('PC result:\n'); print(pc_result$skel) }

        Z <- Z + enforce.rule1( pc_result$skel, which( e$e == 1 ),
                                sepset=pc_result$sepset, sign=pc_result$sign,
                                verbose=verbose )$Z
      }

      if ( any(rules == 2 ) ) {
        Z <- Z + enforce.rule2( e$Cx, which( e$e == 1 ), which( e$e == 0 ),
                                pind=R$pind, pdep=R$pdep, verbose=verbose )$Z
      }

      if ( any(rules == 3 ) ) {
        Z <- Z + enforce.rule3( e$Cx, which( e$e == 1 ), which( e$e == 0 ),
                                pind=R$pind, pdep=R$pdep, verbose=verbose )$Z
      }

      #adding here the zeros
      for ( i in 1:nrow(Z) ) {
        for ( j in (1:ncol(Z))[-i] ) {
          if ( Z[i,j] > 0 ) {
            eqs<-addZeroEquation( eqs, i, j, w=Z[i,j] )
          }
        }
      }

      zero <- zero | Z > 0
      if ( any( !truezeros & zero & mask ) ) {
        R$extra_zero<-max(R$extra_zero,max(abs(C$B[!truezeros & zero & mask])) )
      }

      if ( verbose > 0 ) {
        cat('Largest element put to zero:',R$extra_zero,'\n')

        cat('Unfound zeros:', sum( truezeros & !zero & mask), '\n' )
        cat('Extra zeros:', sum( !truezeros & zero & mask), '\n' )
        cat('Correct zeros:', sum( truezeros & zero & mask), '\n' )
        cat('Correct nonzeros:', sum( !truezeros & !zero & mask), '\n' )
      }
    }

    u<-undetermined( eqs ) #finding which elements are still undetermined

    if (verbose > 0 ) { cat('PCs sat before:',sum( eqs$sat & mask),'\n') }
    if (verbose > 0 ) { cat('Determined:',sum( !u & mask ),'\n'); }

    #calculating here the solution using pseudoinverse
    b<-mpinv(eqs$K) %*% eqs$k

    #also estimating the rank for plots later on
    rank<-qr(eqs$K)$rank

    B<- b.to.B(b, eqs$P, eqs$n )

    if ( any( !u & !eqs$sat & mask ) ) {
      R$extra_determined <-max(c( R$extra_determined, max(abs(B - C$B)[!u & !eqs$sat & mask]) ) )
      #if ( R$extra_determined > 1 ) browser() 
    }
    if ( verbose > 0 ) cat('Largest error put determined:',R$extra_determined,'\n')

    if (verbose > 0 ) cat('Adding determined to satisfied...\n');
    eqs$sat <- eqs$sat | (!u & mask )
    if (verbose > 0 ) { cat('PCs sat after:',sum( eqs$sat & mask),'\n') }

    R$PCs<-c(R$PCs,sum(eqs$sat & mask))
    R$Rs<-c(R$Rs,rank)

    inferred<-all( !mask | eqs$sat ) #this checks whether pair condition is satisfied for all pairs

    if (verbose > 0 ) cat('----------------------------------------------\n');
  }

  R$B<-B
  R$nexp<-ei

  cat( 'Error:',max(abs(C$B - R$B ) ),'- nexp:',R$nexp,'\n' )

  invisible(R)
}

